/* 
 * Copyright (C) 2008 Alan Ross, Tojio GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * 
 * Note: This application aims at being the Mac OS X or Linux version of the Windows server,
 * providing data for the WiiFlash 0.4 library. As the source for the Windows server version 
 * was not open when working on this little application the probability of it some how
 * working differently than the windows version is quite high...
 * 
 * To compile properly this software requires following .jar files:
 * 
 * 	   Bluecove (v 2.0.2) http://code.google.com/p/bluecove/
 * 	   WiiRemoteJ (v 1.3) http://www.wiili.org/WiiremoteJ
 * 
 * TO DO:
 * - Send extensions input (is 100% 'guessplimented': don't own an extension and never tested or implemented correctly)
 * - Get correct values for buttons (Classic controller & Nunchuck)
 * - exchange the 'apple inc.' lines of code, to make application runnable on other platforms (WiiFlashServerJ, Line 134)
 * - Testing with more than one Wiimote (sorry haven't got more)
 * - Tidy up this code mess a little
 * - ... so much to do, so little time
*/



package com.tojio.wiiflashserverj;

//import regular java
import javax.swing.*;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;

import java.awt.Font;
import java.awt.Point;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.WindowListener;
import java.awt.event.WindowEvent;
import java.net.ServerSocket;
import java.util.ArrayList;
import java.util.Timer;
import java.util.TimerTask;
import java.io.IOException;

//import WiiremoteJ
import wiiremotej.*;

//import Apple library (just used when closing app, see initUI() )
import com.apple.eawt.*;


public class WiiFlashServerJ
{
	//Application & UI
	private JFrame f;
	private JButton bWii;
	private JButton bLog;
	private JTextArea log;
	private JScrollPane sLog;
	private AboutBox aboutBox;
	
	private boolean showLog;
	
	// button text
	private final String CLASS_NAME = "[WiiFlashServerJ 0.4]";
	private final String WII_CONNECT = "Discover Wiimote(s)";
	private final String WII_DISCONNECT = "Disconnect Wiimote(s)";
	private final String WII_SEARCHING = "Searching...";
	private final String LOG_ENABLE = "Enable Log";
	private final String LOG_DISABLE = "Disable Log";
	
	//Server
	private boolean listen = true;
	private Server server;
	private final int PORT = 0x4a54;	// default WiiFlash port
	
	//List of all connected WiiRemotes
	private ArrayList<WiiMoteAdapter> wiiremotes;
	
	/**
	 * Creates a new WiiFlashServerJ
	 */
	public WiiFlashServerJ()
	{
		wiiremotes = new ArrayList<WiiMoteAdapter>();
		
		showLog = true; 
		
		//WiiRemoteJ should log also, (is sent to console)
		WiiRemoteJ.setConsoleLoggingAll();
		
		initUI();
		startServer();	
	}
	
	
	//---------------------------------------------------------------------------
	// UI & APPLICATION
	//---------------------------------------------------------------------------
	
	/**
	 * initialize the user interface of this app.
	 */
	public void initUI()
	{
		//initialize frame
		f = new JFrame(CLASS_NAME);
		f.setSize(710,350);
		f.setResizable(false);
		f.addWindowListener(new WindowListener()
		{
	    	public void windowOpened(WindowEvent e){}
	    	public void windowClosing(WindowEvent e){ handleQuitAppication(); }
	    	public void windowClosed(WindowEvent e){ }
	    	public void windowIconified(WindowEvent e){}
	    	public void windowDeiconified(WindowEvent e){}
	    	public void windowActivated(WindowEvent e){}
	    	public void windowDeactivated(WindowEvent e){}
	    });
		
		//Apple Inc. specific stuff. Its used when app is quitting (by shortcut, menu or window..) (!)
	    Application fApp = Application.getApplication();
		fApp.addApplicationListener(new com.apple.eawt.ApplicationAdapter()
		{ 
			public void handleQuit(ApplicationEvent e){ handleQuitAppication(); }
			public void handleAbout(ApplicationEvent e) {
					if (aboutBox == null) { aboutBox = new AboutBox(); }
                    aboutBox.setResizable(false);
					aboutBox.setVisible(true);
                    e.setHandled(true);
			}
		});
		
		//add button Wii (Discover & Disconnect)
		bWii = new JButton(WII_CONNECT);
		bWii.setBounds(10,10,190,30);
		bWii.addActionListener(new ActionListener()
		{ public void actionPerformed(ActionEvent e){ handleWii();  }});
		
		//add button Disconnect
		bLog = new JButton(LOG_DISABLE);
		bLog.setBounds(10,50,190,30);
		bLog.addActionListener(new ActionListener()
		{ public void actionPerformed(ActionEvent e){ handleLog(); }});
		
		//add debug / log text area
		log = new JTextArea();
		log.setFont(new Font("SansSerif", Font.PLAIN, 12));
		sLog = new JScrollPane(log);
		sLog.setBounds(210,10,490,310);
		sLog.setBorder(new CompoundBorder(new EmptyBorder(0,0,0,0) , new EtchedBorder()));
		
		f.getContentPane().setLayout(null);
		f.getContentPane().add(bWii);
		f.getContentPane().add(bLog);
		f.getContentPane().add(sLog);
		
		//now set frame visible
		f.setVisible(true);
	    f.setEnabled(true);
	}
	
	
	/**
	 * called when connect / disconnect button is clicked.
	 * look for WiiRemotes, disconnect WiiRemotes.
	 */
	public void handleWii() 
	{	
		if(wiiremotes.size() == 0)
		{
			log("called search for WiiRemote(s)");
			discoverWii();
		}
		else
		{
			log("called disconnect WiiRemote(s)");
			disconnectWii();
		}
	}
	
	/**
	 * called when log button was clicked. enable, disable log. 
	 */
	public void handleLog() 
	{	
		showLog = !showLog;
		if(showLog){ log("Log enabled"); bLog.setText(LOG_DISABLE); }
		else{ log("Log disabled"); bLog.setText(LOG_ENABLE); }
	}
	
	
	/**
	 * Important: Disconnect all Wiis so next time you try to connect you don't get any errors.
	 * well, you should get a few less.
	 * If WiiRemotes don't seem to be able to connect (on Mac OS X), go to 
	 * 'System Preferences' > 'Bluetooth' and remote your WiiRemotes from the list. 
	 */
	public void handleQuitAppication() 
	{	
		disconnectWii();
		stopServer();
		System.out.println(CLASS_NAME +  " quit");
		System.exit(0);
	}

	
	/**
	 * write a message to the log
	 */
	public void log(String msg)
	{
		if(showLog)
		{
			// Using JTextPane
			//log.setContentType("text/html");
			//log.setText(log.getText() + "<span style=\"font-family: Helvetica, sans-serif; font-size:10px; font-weight:bold;\">" + CLASS_NAME +  "</span>" +
			//		"<span style=\"font-family: Helvetica, sans-serif; font-size:10px; font-weight:normal;\"> " + msg + "</span>\n");
			
			sLog.getViewport().setViewPosition(new Point(0,log.getLineCount()*100));
			log.append(CLASS_NAME +  " " + msg + "\n");
		}
	}
	
	
	
	/**
	 * write a message and the name of where message is coming from to the log
	 */
	public void log(String name, String msg)
	{
		if(showLog)
		{
			sLog.getViewport().setViewPosition(new Point(0,log.getLineCount()*100));
			log.append(name +  " " + msg + "\n");
		}
	}
	
	
	
	
	//---------------------------------------------------------------------------
	// SERVER
	//---------------------------------------------------------------------------
	
	
	/**
	 * Start the Server. Waits till a client is found, then makes instance of Server
	 * so more then one client can connect.
	 * This is also the place where a few seconds of thought should be regarded as
	 * helpful. If one client gives WiiRemote assigns an id and the other client does so
	 * too, but with an other id.. or what happens of a client request setting the LED 
	 * or rumble... but the other client does not... hmm.
	 */
	public void startServer()
	{
		log("starting Server on port: " + PORT);
		
		try
		{
			ServerSocket serverSocket = null;
			
			try{ serverSocket = new ServerSocket(PORT); } 
			catch (Exception e)
			{ 
				log("could NOT listen on port: " + PORT); 
				log("Try:","Quit other versions of this app that are running at the same time");
				log("Try:","Restart this application");
				listen = false; 
				}
			
			while (listen)
			{
				log("waiting for client on port: " + PORT);
				    	
				server = new Server(this, serverSocket.accept());
				server.start();
			}
			
			serverSocket.close();
		}
	    catch (IOException e){ log("Server IOException: " + e.getMessage() ); }
	}
	
	
	
	
	/**
	 * Stop the server.
	 */
	public void stopServer()
	{
		log("shutdown Server");
		
		listen = false;
		if(server != null) server.quit();
	}
	
	
	/**
	 * Return true if server is available and running, false if not.
	 */
	public boolean serverIsRunning()
	{
		if(server != null) return server.isRunning();
		else return false;
	}
	


	/**
	 * write data do socket (send WiiRemote data to the server to write to socket / WiiFlash)
	 */
	public void sendRemoteData(WiiMoteData remoteData)
	{
		if(server != null)
		{ 
			//log("receiving WiiRemote data & server sending");
			server.sendRemoteData(remoteData);
		}
		//else{ log("receiving WiiRemote data"); }
	}
	
	
	
	//---------------------------------------------------------------------------
	// WIIREMOTE
    //---------------------------------------------------------------------------
	
	/**
	 * Look for WiiRemotes. this means WiiRemotes that have their
	 * 1 & 2 Button pressed.
	 */
	public void discoverWii() 
	{
		log("Searching. Press Button 1 & 2 on WiiRemote");
		bWii.setText(WII_SEARCHING);
		
		WiiRemoteJ.findRemotes(new WiiMoteFinder(this));
		
		//wait 10 seconds, if nothing connected, or error, reset search button...
		Timer t = new Timer();
		t.schedule(new TimerTask()
		{ 
			public void run()
			{ 
				if(wiiremotes.size()==0)
				{
					log("Stop searching for WiiRemote(s)");
					log("(!)","Your WiiRemote should have been found?");
					log("-> ","Press Button 1 & 2 on your WiiRemote and try again");
					log("-> ","OS X: Go to 'System Preferences' and delete all Remotes in 'Bluetooth'");
					log("-> ","Quit other versions of this app that are running at the same time");
					log("-> ","Quit other applications that are connected to your Wiiremote");
					log("-> ","Other devices trying to connect with the Wiiremote(s)?");
					log("-> ","Restart this application");
					log("-> ","Hmm. Any other ideas?");
					
					disconnectWii();
				}
			} 
		}, 10000);
    }
	
	
	
	/**
	 * Disconnect all WiiRemotes that are connected to WiiFlashServerJ.
	 * Should be called when quitting the application, to free all connected
	 * WiiRemotes. More info to why we want to do this: handleQuitAppication()
	 */
	public void disconnectWii() 
	{
		log("disconnect WiiRemote(s)");
		
		WiiRemoteJ.stopFind();
		
		for(int i = 0; i< wiiremotes.size(); i++)
		{
			((WiiMoteAdapter) wiiremotes.get(i)).disconnect();
			log("disconnect WiiRemote [" + i+"]");
		}
		
		wiiremotes.clear();
		bWii.setText(WII_CONNECT);
	}
	
	
	
	
	/**
	 * Add a WiiMoteAdapter to WiiFlashServerJ, so the server can send its data.
	 */
	public int registerRemote(WiiMoteAdapter wma)
	{
		if(wiiremotes.size() < 4) //just allow 4 WiiRemotes to be connected
		{
			log("register WiiRemote [" + wiiremotes.size() + "]:"+ wma.getBluetoothAddress());
			wiiremotes.add(wma);
			bWii.setText(WII_DISCONNECT);
			return wiiremotes.size()-1; //size()-1 = WiiRemote id
		}
		else
		{
			log("could NOT register WiiRemote [" + (wiiremotes.size()-1) + "]:"+ wma.getBluetoothAddress() + " (to many connected)");
			return -1;
		}
	
		
	}
	
	
	/**
	 * Add a WiiMoteAdapter to WiiFlashServerJ, so the server can send its data.
	 */
	public void removeRemote(WiiMoteAdapter wma)
	{
		log("remove WiiRemote (NOT IMPLEMENTED YET)");
	}
	
	
	/**
	 * Return the amount of WiiRemotes currently handled by the WiiFlashServerJ
	 */
	public int getRemoteCount()
	{
		return wiiremotes.size();
	}
	

	/**
	 * Set the LED Lights of a given WiiRemote
	 */
	public void setLEDLights(int remoteID, int lightID)
	{
		log("client request: Wii["+remoteID+"] set LED" + lightID);
		
		for(int i=0; i< wiiremotes.size(); i++)
		{
			if(wiiremotes.get(i).getID() == remoteID)
			{ 
				if(0 <= lightID && lightID <=3 )
				{
					wiiremotes.get(remoteID).setLEDLights(lightID);
				}
			}
		}
	}
	
	/**
	 * Start vibrating a WiiRemote, given by its ID
	 */
	public void startVibrating(int remoteID)
	{
		log("client request: Wii["+remoteID+"] start vibrating");
		
		for(int i=0; i< wiiremotes.size(); i++)
		{
			if(wiiremotes.get(i).getID() == remoteID)
			{ wiiremotes.get(i).startVibrating(); }
		}
	}
	
	
	/**
	 * Stop vibrating a WiiRemote, given by its ID
	 */
	public void stopVibrating(int remoteID)
	{
		log("client request: Wii["+remoteID+"] stop vibrating");
		for(int i=0; i< wiiremotes.size(); i++)
		{
			if(wiiremotes.get(i).getID() == remoteID)
			{ wiiremotes.get(i).stopVibrating(); }
		}
	}
	
	
	
	
	/**
	 * Starts this application
	 */
	public static void main(String args[])
	{
		new WiiFlashServerJ();
	}

}
